@@ -1,6 +1,6 @@
 /*
  * LibSylph -- E-Mail client library
- * Copyright (C) 1999-2006 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2007 Hiroyuki Yamamoto
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -398,18 +398,21 @@ gpointer my_memmem(gconstpointer haystack, size_t haystacklen,
 	const gchar *haystack_ = (const gchar *)haystack;
 	const gchar *needle_ = (const gchar *)needle;
 	const gchar *haystack_cur = (const gchar *)haystack;
+	size_t haystack_left = haystacklen;
 
 	if (needlelen == 1)
 		return memchr(haystack_, *needle_, haystacklen);
 
-	while ((haystack_cur = memchr(haystack_cur, *needle_, haystacklen))
+	while ((haystack_cur = memchr(haystack_cur, *needle_, haystack_left))
 	       != NULL) {
 		if (haystacklen - (haystack_cur - haystack_) < needlelen)
 			break;
 		if (memcmp(haystack_cur + 1, needle_ + 1, needlelen - 1) == 0)
 			return (gpointer)haystack_cur;
-		else
+		else {
 			haystack_cur++;
+			haystack_left = haystacklen - (haystack_cur - haystack_);
+		}
 	}
 
 	return NULL;
@@ -468,13 +471,19 @@ gint str_find_format_times(const gchar *haystack, gchar ch)
 gboolean is_next_nonascii(const gchar *s)
 {
 	const gchar *p;
+	gboolean in_quote = FALSE;
 
 	/* skip head space */
 	for (p = s; *p != '\0' && g_ascii_isspace(*p); p++)
 		;
-	for (; *p != '\0' && !g_ascii_isspace(*p); p++) {
-		if (*(guchar *)p > 127 || *(guchar *)p < 32)
+	while (*p != '\0') {
+		if (!in_quote && g_ascii_isspace(*p))
+			break;
+		if (*p == '"')
+			in_quote ^= TRUE;
+		else if (*(guchar *)p > 127 || *(guchar *)p < 32)
 			return TRUE;
+		++p;
 	}
 
 	return FALSE;
@@ -482,12 +491,18 @@ gboolean is_next_nonascii(const gchar *s)
 
 gint get_next_word_len(const gchar *s)
 {
-	gint len = 0;
+	const gchar *p = s;
+	gboolean in_quote = FALSE;
 
-	for (; *s != '\0' && !g_ascii_isspace(*s); s++, len++)
-		;
+	while (*p != '\0') {
+		if (!in_quote && g_ascii_isspace(*p))
+			break;
+		if (*p == '"')
+			in_quote ^= TRUE;
+		++p;
+	}
 
-	return len;
+	return p - s;
 }
 
 /* compare subjects */
@@ -639,6 +654,36 @@ void extract_parenthesis(gchar *str, gchar op, gchar cl)
 	*destp = '\0';
 }
 
+void extract_parenthesis_with_escape(gchar *str, gchar op, gchar cl)
+{
+	register gchar *srcp, *destp;
+	gint in_brace;
+
+	srcp = destp = str;
+
+	while ((srcp = strchr(srcp, op))) {
+		if (destp > str)
+			*destp++ = ' ';
+		++srcp;
+		in_brace = 1;
+		while (*srcp) {
+			if (*srcp == op)
+				in_brace++;
+			else if (*srcp == cl)
+				in_brace--;
+
+			if (in_brace == 0)
+				break;
+
+			if (*srcp == '\\' && *(srcp + 1) != '\0')
+				++srcp;
+
+			*destp++ = *srcp++;
+		}
+	}
+	*destp = '\0';
+}
+
 void extract_parenthesis_with_skip_quote(gchar *str, gchar quote_chr,
 					 gchar op, gchar cl)
 {
@@ -700,6 +745,25 @@ void extract_quote(gchar *str, gchar quote_chr)
 	}
 }
 
+void extract_quote_with_escape(gchar *str, gchar quote_chr)
+{
+	register gchar *sp, *dp;
+
+	if ((sp = strchr(str, quote_chr))) {
+		dp = sp;
+		++sp;
+		while (*sp) {
+			if (*sp == quote_chr)
+				break;
+			else if (*sp == '\\' && *(sp + 1) != '\0')
+				++sp;
+
+			*dp++ = *sp++;
+		}
+		*dp = '\0';
+	}
+}
+
 void eliminate_address_comment(gchar *str)
 {
 	register gchar *srcp, *destp;
@@ -1007,7 +1071,7 @@ GList *add_history(GList *list, const gchar *str)
 		last = g_list_last(list);
 		if (last) {
 			g_free(last->data);
-			g_list_remove(list, last->data);
+			list = g_list_remove(list, last->data);
 		}
 	}
 
@@ -1342,7 +1406,6 @@ gchar **strsplit_parenthesis(const gchar *str, gchar op, gchar cl,
 	str_array = g_new(gchar*, n);
 
 	i = n - 1;
-
 	str_array[i--] = NULL;
 	for (slist = string_list; slist; slist = slist->next)
 		str_array[i--] = slist->data;
@@ -1402,7 +1465,79 @@ gchar **strsplit_with_quote(const gchar *str, const gchar *delim,
 	str_array = g_new(gchar*, n);
 
 	i = n - 1;
+	str_array[i--] = NULL;
+	for (slist = string_list; slist; slist = slist->next)
+		str_array[i--] = slist->data;
 
+	g_slist_free(string_list);
+
+	return str_array;
+}
+
+gchar **strsplit_csv(const gchar *str, gchar delim, gint max_tokens)
+{
+	GSList *string_list = NULL, *slist;
+	gchar **str_array, *s, *new_str;
+	gchar *tmp, *tmpp, *p;
+	guint i, n = 1, len;
+
+	g_return_val_if_fail(str != NULL, NULL);
+
+	if (max_tokens < 1)
+		max_tokens = G_MAXINT;
+
+	s = strchr_with_skip_quote(str, '"', delim);
+	if (s) {
+		do {
+			len = s - str;
+			tmpp = tmp = g_strndup(str, len);
+
+			if (tmp[0] == '"' && tmp[len - 1] == tmp[0]) {
+				tmp[len - 1] = '\0';
+				++tmpp;
+				p = new_str = g_malloc(len - 1);
+				while (*tmpp) {
+					if (*tmpp == '"' && *(tmpp + 1) == '"')
+						++tmpp;
+					*p++ = *tmpp++;
+				}
+				*p = '\0';
+				g_free(tmp);
+			} else
+				new_str = tmp;
+
+			string_list = g_slist_prepend(string_list, new_str);
+			n++;
+			str = s + 1;
+			s = strchr_with_skip_quote(str, '"', delim);
+		} while (--max_tokens && s);
+	}
+
+	if (*str) {
+		len = strlen(str);
+		tmpp = tmp = g_strdup(str);
+
+		if (tmp[0] == '"' && tmp[len - 1] == tmp[0]) {
+			tmp[len - 1] = '\0';
+			++tmpp;
+			p = new_str = g_malloc(len - 1);
+			while (*tmpp) {
+				if (*tmpp == '"' && *(tmpp + 1) == '"')
+					++tmpp;
+				*p++ = *tmpp++;
+			}
+			*p = '\0';
+			g_free(tmp);
+		} else
+			new_str = tmp;
+
+		string_list = g_slist_prepend(string_list, new_str);
+		n++;
+	}
+
+	str_array = g_new(gchar*, n);
+
+	i = n - 1;
 	str_array[i--] = NULL;
 	for (slist = string_list; slist; slist = slist->next)
 		str_array[i--] = slist->data;
@@ -1711,6 +1846,26 @@ gchar *uriencode_for_filename(const gchar *filename)
 	return enc;
 }
 
+gchar *uriencode_for_mailto(const gchar *mailto)
+{
+	const gchar *p = mailto;
+	gchar *enc, *outp;
+
+	outp = enc = g_malloc(strlen(mailto) * 3 + 1);
+
+	for (p = mailto; *p != '\0'; p++) {
+		if (*p == '+') {
+			*outp++ = '%';
+			get_hex_str(outp, *p);
+			outp += 2;
+		} else
+			*outp++ = *p;
+	}
+
+	*outp = '\0';
+	return enc;
+}
+
 gint scan_mailto_url(const gchar *mailto, gchar **to, gchar **cc, gchar **bcc,
 		     gchar **subject, gchar **body)
 {
@@ -1816,6 +1971,28 @@ const gchar *get_startup_dir(void)
 	return startup_dir;
 }
 
+#ifdef G_OS_WIN32
+static gchar *get_win32_special_folder_path(gint nfolder)
+{
+	gchar *folder = NULL;
+	HRESULT hr;
+
+	if (G_WIN32_HAVE_WIDECHAR_API()) {
+		wchar_t path[MAX_PATH + 1];
+		hr = SHGetFolderPathW(NULL, nfolder, NULL, 0, path);
+		if (hr == S_OK)
+			folder = g_utf16_to_utf8(path, -1, NULL, NULL, NULL);
+	} else {
+		gchar path[MAX_PATH + 1];
+		hr = SHGetFolderPathA(NULL, nfolder, NULL, 0, path);
+		if (hr == S_OK)
+			folder = g_locale_to_utf8(path, -1, NULL, NULL, NULL);
+	}
+
+	return folder;
+}
+#endif
+
 const gchar *get_home_dir(void)
 {
 #ifdef G_OS_WIN32
@@ -1840,23 +2017,7 @@ const gchar *get_document_dir(void)
 	HRESULT hr;
 
 	if (!document_dir) {
-		if (G_WIN32_HAVE_WIDECHAR_API()) {
-			wchar_t path[MAX_PATH + 1];
-			hr = SHGetFolderPathW(NULL, CSIDL_PERSONAL, NULL, 0,
-					      path);
-			if (hr == S_OK) {
-				document_dir = g_utf16_to_utf8
-					(path, -1, NULL, NULL, NULL);
-			}
-		} else {
-			gchar path[MAX_PATH + 1];
-			hr = SHGetFolderPathA(NULL, CSIDL_PERSONAL, NULL, 0,
-					      path);
-			if (hr == S_OK) {
-				document_dir = g_locale_to_utf8
-					(path, -1, NULL, NULL, NULL);
-			}
-		}
+		document_dir = get_win32_special_folder_path(CSIDL_PERSONAL);
 		if (!document_dir)
 			document_dir = get_home_dir();
 	}
@@ -1871,15 +2032,16 @@ const gchar *get_rc_dir(void)
 {
 	if (!rc_dir) {
 #ifdef G_OS_WIN32
-		const gchar *appdata;
+		gchar *appdata;
 
-		appdata = g_getenv("APPDATA");
+		appdata = get_win32_special_folder_path(CSIDL_APPDATA);
 		if (appdata)
 			rc_dir = g_strconcat(appdata, G_DIR_SEPARATOR_S,
 					     RC_DIR, NULL);
 		else
 			rc_dir = g_strconcat(get_home_dir(), G_DIR_SEPARATOR_S,
 					     RC_DIR, NULL);
+		g_free(appdata);
 #else
 		rc_dir = g_strconcat(get_home_dir(), G_DIR_SEPARATOR_S,
 				     RC_DIR, NULL);
@@ -1983,7 +2145,6 @@ gchar *get_tmp_file(void)
 
 const gchar *get_domain_name(void)
 {
-#ifdef G_OS_UNIX
 	static gchar *domain_name = NULL;
 
 	if (!domain_name) {
@@ -2004,12 +2165,14 @@ const gchar *get_domain_name(void)
 		}
 
 		debug_print("domain name = %s\n", domain_name);
+		if (is_next_nonascii(domain_name)) {
+			g_warning("invalid domain name: %s\n", domain_name);
+			g_free(domain_name);
+			domain_name = "unknown";
+		}
 	}
 
 	return domain_name;
-#else
-	return "unknown";
-#endif
 }
 
 off_t get_file_size(const gchar *file)
@@ -2079,18 +2242,21 @@ off_t get_left_file_size(FILE *fp)
 
 gboolean file_exist(const gchar *file, gboolean allow_fifo)
 {
-	struct stat s;
-
 	if (file == NULL)
 		return FALSE;
 
-	if (g_stat(file, &s) < 0) {
-		if (ENOENT != errno) FILE_OP_ERROR(file, "stat");
-		return FALSE;
-	}
+	if (allow_fifo) {
+		struct stat s;
 
-	if (S_ISREG(s.st_mode) || (allow_fifo && S_ISFIFO(s.st_mode)))
-		return TRUE;
+		if (g_stat(file, &s) < 0) {
+			if (ENOENT != errno) FILE_OP_ERROR(file, "stat");
+			return FALSE;
+		}
+		if (S_ISREG(s.st_mode) || S_ISFIFO(s.st_mode))
+			return TRUE;
+	} else {
+		return g_file_test(file, G_FILE_TEST_IS_REGULAR);
+	}
 
 	return FALSE;
 }
@@ -2404,8 +2570,15 @@ static gint remove_dir_recursive_real(const gchar *dir)
 	g_free(prev_dir);
 
 	if (g_rmdir(dir) < 0) {
-		FILE_OP_ERROR(dir, "rmdir");
-		return -1;
+		if (ENOTDIR == errno) {
+			if (g_unlink(dir) < 0) {
+				FILE_OP_ERROR(dir, "unlink");
+				return -1;
+			}
+		} else {
+			FILE_OP_ERROR(dir, "rmdir");
+			return -1;
+		}
 	}
 
 	return 0;
@@ -2445,7 +2618,7 @@ leave:
 
 gint rename_force(const gchar *oldpath, const gchar *newpath)
 {
-#ifndef G_OS_UNIX
+#if !defined(G_OS_UNIX) && !GLIB_CHECK_VERSION(2, 9, 1)
 	if (!is_file_entry_exist(oldpath)) {
 		errno = ENOENT;
 		return -1;
@@ -2460,29 +2633,37 @@ gint rename_force(const gchar *oldpath, const gchar *newpath)
 
 gint copy_file(const gchar *src, const gchar *dest, gboolean keep_backup)
 {
-	FILE *src_fp, *dest_fp;
+	gint srcfd, destfd;
 	gint n_read;
-	gchar buf[BUFSIZ];
+	gchar buf[BUFFSIZE];
 	gchar *dest_bak = NULL;
 	gboolean err = FALSE;
 
-	if ((src_fp = g_fopen(src, "rb")) == NULL) {
-		FILE_OP_ERROR(src, "fopen");
+#ifdef G_OS_WIN32
+	if ((srcfd = g_open(src, O_RDONLY | O_BINARY, 0600)) < 0) {
+#else
+	if ((srcfd = g_open(src, O_RDONLY, 0600)) < 0) {
+#endif
+		FILE_OP_ERROR(src, "open");
 		return -1;
 	}
 	if (is_file_exist(dest)) {
 		dest_bak = g_strconcat(dest, ".bak", NULL);
 		if (rename_force(dest, dest_bak) < 0) {
 			FILE_OP_ERROR(dest, "rename");
-			fclose(src_fp);
+			close(srcfd);
 			g_free(dest_bak);
 			return -1;
 		}
 	}
 
-	if ((dest_fp = g_fopen(dest, "wb")) == NULL) {
-		FILE_OP_ERROR(dest, "fopen");
-		fclose(src_fp);
+#ifdef G_OS_WIN32
+	if ((destfd = g_open(dest, O_WRONLY | O_CREAT | O_BINARY, 0600)) < 0) {
+#else
+	if ((destfd = g_open(dest, O_WRONLY | O_CREAT, 0600)) < 0) {
+#endif
+		FILE_OP_ERROR(dest, "open");
+		close(srcfd);
 		if (dest_bak) {
 			if (rename_force(dest_bak, dest) < 0)
 				FILE_OP_ERROR(dest_bak, "rename");
@@ -2491,37 +2672,33 @@ gint copy_file(const gchar *src, const gchar *dest, gboolean keep_backup)
 		return -1;
 	}
 
-	if (change_file_mode_rw(dest_fp, dest) < 0) {
-		FILE_OP_ERROR(dest, "chmod");
-		g_warning(_("can't change file mode\n"));
-	}
+	while ((n_read = read(srcfd, buf, sizeof(buf))) > 0) {
+		gchar *p = buf;
+		const gchar *endp = buf + n_read;
+		gint n_write;
 
-	while ((n_read = fread(buf, sizeof(gchar), sizeof(buf), src_fp)) > 0) {
-		if (n_read < sizeof(buf) && ferror(src_fp))
-			break;
-		if (fwrite(buf, n_read, 1, dest_fp) < 1) {
-			g_warning(_("writing to %s failed.\n"), dest);
-			fclose(dest_fp);
-			fclose(src_fp);
-			g_unlink(dest);
-			if (dest_bak) {
-				if (rename_force(dest_bak, dest) < 0)
-					FILE_OP_ERROR(dest_bak, "rename");
-				g_free(dest_bak);
+		while (p < endp) {
+			if ((n_write = write(destfd, p, endp - p)) < 0) {
+				g_warning(_("writing to %s failed.\n"), dest);
+				close(destfd);
+				close(srcfd);
+				g_unlink(dest);
+				if (dest_bak) {
+					if (rename_force(dest_bak, dest) < 0)
+						FILE_OP_ERROR(dest_bak, "rename");
+					g_free(dest_bak);
+				}
+				return -1;
 			}
-			return -1;
+			p += n_write;
 		}
 	}
 
-	if (ferror(src_fp)) {
-		FILE_OP_ERROR(src, "fread");
-		err = TRUE;
-	}
-	fclose(src_fp);
-	if (fclose(dest_fp) == EOF) {
-		FILE_OP_ERROR(dest, "fclose");
+	if (close(destfd) < 0) {
+		FILE_OP_ERROR(dest, "close");
 		err = TRUE;
 	}
+	close(srcfd);
 
 	if (err) {
 		g_unlink(dest);
@@ -2575,7 +2752,7 @@ gint copy_dir(const gchar *src, const gchar *dest)
 
 gint move_file(const gchar *src, const gchar *dest, gboolean overwrite)
 {
-	if (overwrite == FALSE && is_file_exist(dest)) {
+	if (overwrite == FALSE && is_file_entry_exist(dest)) {
 		g_warning("move_file(): file %s already exists.", dest);
 		return -1;
 	}
@@ -3152,10 +3329,47 @@ gchar *generate_mime_boundary(const gchar *prefix)
 
 gint change_file_mode_rw(FILE *fp, const gchar *file)
 {
-#if HAVE_FCHMOD
-	return fchmod(fileno(fp), S_IRUSR|S_IWUSR);
+#ifdef G_OS_WIN32
+	DWORD attr;
+	BOOL retval;
+
+	if (G_WIN32_HAVE_WIDECHAR_API()) {
+		wchar_t *wpath;
+
+		wpath = g_utf8_to_utf16(file, -1, NULL, NULL, NULL);
+		if (wpath == NULL)
+			return -1;
+
+		attr = GetFileAttributesW(wpath);
+		retval = SetFileAttributesW
+			(wpath, attr & ~(FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_HIDDEN));
+
+		g_free(wpath);
+	} else {
+		gchar *cp_path;
+
+		cp_path = g_locale_from_utf8(file, -1, NULL, NULL, NULL);
+		if (cp_path == NULL)
+			return -1;
+
+		attr = GetFileAttributesA(cp_path);
+		retval = SetFileAttributesA
+			(cp_path, attr & ~(FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_HIDDEN));
+
+		g_free(cp_path);
+	}
+
+	if (retval)
+		return 0;
+	else
+		return -1;
 #else
-	return g_chmod(file, S_IRUSR|S_IWUSR);
+#if HAVE_FCHMOD
+	if (fp)
+		return fchmod(fileno(fp), S_IRUSR|S_IWUSR);
+	else
+#endif
+		return g_chmod(file, S_IRUSR|S_IWUSR);
 #endif
 }
 
@@ -3988,7 +4202,7 @@ void log_print(const gchar *format, ...)
 	g_vsnprintf(buf + TIME_LEN, BUFFSIZE, format, args);
 	va_end(args);
 
-	if (debug_mode) fputs(buf, stdout);
+	if (debug_mode) g_print("%s", buf);
 	log_print_ui_func(buf);
 	if (log_fp) {
 		fputs(buf, log_fp);
